@@ -66,13 +66,19 @@ def logerr(line):
     if line:
         log.error(line)
 
+# The sh debug logging swamps the log file, so back off.
+def _adjust_sh_log():
+    logger = logging.getLogger(sh.__name__)
+    if logger.getEffectiveLevel() <= logging.DEBUG:
+        logger.setLevel(logging.INFO)
+
 # Commands
 cp = sh.Command('cp').bake(_out=logout, _err=logerr)
 ssh = sh.Command('ssh')
 ssh_keygen = sh.Command('ssh-keygen').bake(_out=logout, _err=logerr)
+qemu_img = sh.Command('qemu-img').bake(_out=logout, _err=logerr)
 virt_builder = sh.Command('virt-builder').bake(_out=logout, _err=logerr)
 virt_install = sh.Command('virt-install').bake(_out=logout, _err=logerr)
-virt_clone = sh.Command('virt-clone').bake(_out=logout, _err=logerr)
 virt_sysprep = sh.Command('virt-sysprep').bake(_out=logout, _err=logerr)
 
 # Avoid writing "domain not found" errors to the console.
@@ -94,6 +100,7 @@ class Settings:
         self.username = self.site.get('username', 'virt')
         self.image_format = self.site.get('image-format', 'qcow2')
         self.dns_domain = self.site.get('dns-domain', '')
+        self.address_source = self.site.get('address-source', 'agent')
 
         # Load template definitions.
         self.templates = self._load('templates', TEMPLATES)
@@ -109,6 +116,7 @@ class Settings:
             self.template_name = template
             self.os_version = self.template.get('os-version')
             self.os_variant = self.template.get('os-variant')
+            self.address_source = self.template.get('address-source', self.address_source)
 
     def _load(self, name, defaults):
         system_file = f'/etc/virt-up/{name}.cfg'
@@ -268,9 +276,16 @@ class Instance:
         self._read_meta()
 
     def _update_meta(self, meta):
+        changed = []
         for key in meta:
-            self.meta[key] = meta[key]
-        self._write_meta()
+            value = meta[key]
+            if not key in self.meta or self.meta[key] != value:
+                self.meta[key] = value
+                changed.append(key)
+        if changed:
+            changed = ', '.join(changed)
+            log.debug(f"Updating metadata fields: {changed}")
+            self._write_meta()
 
     def _read_meta(self):
         try:
@@ -280,8 +295,10 @@ class Instance:
             pass
 
     def _write_meta(self):
+        log.debug(f"Writing metafile '{self.metafile}'.")
         mkdir_p(os.path.dirname(self.metafile))
-        with open(self.metafile, 'w') as fp:
+        flags = os.O_CREAT | os.O_TRUNC | os.O_RDWR
+        with os.fdopen(os.open(self.metafile, flags, 0o600), 'w') as fp:
             json.dump(self.meta, fp, indent=4)
 
     def mac(self):
@@ -397,20 +414,20 @@ class Instance:
                         lines.append(f'    {i:8}  {mac:17}  {aip}')
         return '\n'.join(lines)
 
-    def address(self, source=None):
+    def _address_from_ia(self, source='agent'):
         """
-        Get the public IPv4 address for login.
+        Attempt to get the instance address from the domain interface-addresses.
         """
-        address = self.meta.get('address')
-        if address:
-            return address
-
-        if not self.domain.isActive():
-            self.start()
+        sources = {
+            'agent': libvirt.VIR_DOMAIN_INTERFACE_ADDRESSES_SRC_AGENT,
+            'lease': libvirt.VIR_DOMAIN_INTERFACE_ADDRESSES_SRC_LEASE,
+        }
+        try:
+            source = sources[source]
+        except KeyError:
+            raise ValueError(f"Unsupported interface-address source '{source}.")
 
         log.info(f"Waiting for instance '{self.name}' address.")
-        if source is None:
-            source = libvirt.VIR_DOMAIN_INTERFACE_ADDRESSES_SRC_AGENT
         last_ia_str = None
         addresses = []
         for retries in range(120, -1, -1):
@@ -433,10 +450,60 @@ class Instance:
 
         if not addresses:
             raise LookupError(f"Unable to find address for instance '{self.name}'.")
-        address = addresses[0]
-        self.meta['address'] = address
-        self._write_meta()
+        return addresses[0] # return the first one found.
 
+    def _arp_table(self):
+        """
+        Retrieve the arp table.
+        """
+        arp = {}
+        with open('/proc/net/arp') as fp:
+            for line in fp:
+                if line.startswith('IP address'):
+                    continue
+                address, _, _, mac, _, _ = line.split()
+                arp[mac] = address
+        return arp
+
+    def _address_from_arp(self):
+        """
+        Attempt to retreive the instance address from the arp cache.
+        """
+        address = None
+        mac = self.mac()
+        for retries in range(120, -1, -1):
+            arp = self._arp_table()
+            address = arp.get(mac)
+            if address:
+                break
+            if retries > 0:
+                suffix = 'ies' if retries > 1 else 'y'
+                log.debug(f"Waiting for instance '{self.name}' address in arp cache; {retries} retr{suffix} left.")
+                time.sleep(2)
+        return address
+
+    def address(self):
+        """
+        Get the public IPv4 address for login.
+        """
+        address = self.meta.get('address')
+        if address:
+            return address
+
+        if not self.domain.isActive():
+            self.start()
+
+        address_source = self.meta.get('address_source', 'agent')
+        if address_source == 'agent':
+            address = self._address_from_ia(source='agent')
+        elif address_source == 'lease':
+            address = self._address_from_ia(source='lease')
+        elif address_source == 'arp':
+            address = self._address_from_arp()
+        else:
+            raise ValueError(f"Invalid address_source '{address_source}' in instance '{self.name}'.")
+
+        self._update_meta({'address': address})
         log.info(f"Instance '{self.name}' has address '{address}'.")
         return address
 
@@ -451,8 +518,9 @@ class Instance:
                 s.settimeout(2)
                 s.connect((address, int(port)))
                 return True
-            except:
-                pass
+            except Exception as e:
+                if str(e) == "[Errno 111] Connection refused":
+                    time.sleep(2) 
             finally:
                 try:
                     s.shutdown(socket.SHUT_RDWR)
@@ -472,7 +540,7 @@ class Instance:
                 name = domain.name()
                 metafile = f'{xdg_data_home}/virt-up/instance/{name}.json'
                 try:
-                    with open(metafile) as fp:
+                    with open(metafile, 'r') as fp:
                         meta = json.load(fp)
                     if clones_only:
                         if 'cloned' in meta:
@@ -483,18 +551,9 @@ class Instance:
                     pass
 
     @classmethod
-    def images(cls):
+    def _domain_exists(cls, name):
         """
-        List the available build() image names.
-        """
-        settings = Settings()
-        for name, values in settings.images.items():
-            yield (name, values.get('desc', ''), values.get('arch', ''))
-
-    @classmethod
-    def exists(cls, name):
-        """
-        Returns true if domain already exists.
+        Returns true if the domain already exists.
         """
         assert(name)
         domain = None
@@ -506,11 +565,24 @@ class Instance:
                     raise e
         return (not domain is None)
 
+    @classmethod
+    def exists(cls, name):
+        """
+        Returns true if the instance already exists, that is
+        the domain and metadata file both exist.
+        """
+        metafile = f'{xdg_data_home}/virt-up/instance/{name}.json'
+        return os.path.exists(metafile) and cls._domain_exists(name)
+
     @classmethod
     def build(cls,
               name,
               template=None, # defaults to <name>
               prefix='',
+              settings=None,
+              root_password=None,
+              user=None,
+              password=None,
               memory=512,
               size=None,
               vcpus=1,
@@ -518,6 +590,7 @@ class Instance:
         """
         Build an instance with virt-builder and virt-install.
         """
+        _adjust_sh_log()
         if name is None:
             raise ValueError('<name> is required.')
         if not template:
@@ -529,9 +602,13 @@ class Instance:
             log.info(f"Instance '{name}' already exists.")
             return Instance(name)
 
-        settings = Settings(template)
+        if settings is None:
+            settings = Settings(template)
         maddrs = MacAddresses()
         path = query_storage_pool(settings.pool)
+        if not os.access(path, os.R_OK | os.W_OK):
+            raise PermissionError(f"Read and write access is required for path '{path}'.")
+
         image = f'{path}/{name}.{settings.image_format}'
 
         # Sanity checks.
@@ -539,17 +616,21 @@ class Instance:
             raise LookupError(f"virt-builder <os_version> is not defined for '{template}'.")
         if not settings.os_variant:
             raise LookupError(f"virt-install <os_variant> is not defined for '{template}'.")
-
-        # Remove old image file, if one exists.
-        rm_f(image)
+        if cls._domain_exists(name):
+            raise FileExistsError(f"Domain '{name}' without metadata already exists.")
+        if os.path.exists(image):
+            raise FileExistsError(f"Image file '{image}' already exists.")
 
         # Generate the user creditials for login.
-        user = settings.username
-        root_creds = Creds('root')
-        user_creds = Creds(user)
+        if not user:
+            user = settings.username
+        root_creds = Creds('root', password=root_password)
+        user_creds = Creds(user, password=password)
 
         # Setup virt-builder arguments.
         extra_args = settings.extra_args('virt-builder')
+        if size:
+            extra_args.extend(['--size', size])
 
         log.info(f"Building image file '{image}'.")
         virt_builder(
@@ -558,7 +639,8 @@ class Instance:
             '--format', settings.image_format,
             '--root-password', f'password:{root_creds.password}',
             '--run-command', 'ssh-keygen -A',
-            '--run-command', f'id -u {user} || useradd -m -s /bin/bash {user}',
+            '--run-command', f'id -u {user} || useradd -m -s /bin/bash -p "" {user}',
+            '--run-command', f'test -x /sbin/chpasswd && (echo "{user}:{user_creds.password}" | /sbin/chpasswd)',
             '--ssh-inject', f'{user}:file:{user_creds.ssh_identity}.pub',
             '--run-command', 'mkdir -p /etc/sudoers.d',
             '--write',  f'/etc/sudoers.d/99-sna-devlab:{user} ALL=(ALL) NOPASSWD: ALL',
@@ -598,44 +680,63 @@ class Instance:
             'graphics': graphics,
             'root': vars(root_creds),
             'user': vars(user_creds),
+            'address_source': settings.address_source,
         }
+        if size:
+            meta['size'] = size
         instance = Instance(name, meta=meta)
         maddrs.update(name, instance.mac())
         instance.address() # Wait for address to be assigned.
 
         return instance
 
-    def clone(self, target, memory=None, size=None, vcpus=None, graphics=None):
+    def clone(self,
+            target,
+            settings=None,
+            memory=None,
+            size=None,
+            vcpus=None,
+            graphics=None,
+            **kwargs):
         """
         Clone this instance to a new target instance.
 
         This instance will be stopped if it is running. The image will
         be cloned and virt-sysprep'd for the new target instance.
         """
+        _adjust_sh_log()
         assert(target)
         if self.exists(target):
             log.info(f"Target instance '{target}' already exists.")
             return Instance(target)
 
+        if Instance._domain_exists(target):
+            raise FileExistsError(f"Domain '{target}' without metadata already exists.")
+
         # Required meta data elements needed to clone.
         for element in ('os_version', 'os_variant', 'disk'):
             if not element in self.meta:
                 raise LookupError(f"Element '{element}' is missing in '{self.name}' meta data.")
 
-        settings = Settings(self.meta['template'])
+        if settings is None:
+            settings = Settings(self.meta['template'])
         maddrs = MacAddresses()
         path = query_storage_pool(settings.pool)
+        if not os.access(path, os.R_OK | os.W_OK):
+            raise PermissionError(f"Read and write access is required for path '{path}'.")
 
-        # Ensure we are stopped.
-        self.stop()
-
-        # Setup cp arguments.
+        # Clone the image.
         source_image = self.meta['disk']
         target_image = f'{path}/{target}.{settings.image_format}'
-        extra_args = settings.extra_args('cp')
-
+        if os.path.exists(target_image):
+            raise FileExistsError(f"Image file '{target_image}' already exists.")
+        self.stop()  # Ensure we are stopped before cloning.
         log.info(f"Cloning '{source_image}' to '{target_image}'.")
-        cp(*extra_args, source_image, target_image)
+        if settings.image_format == 'qcow2':
+            qemu_img.create('-f', 'qcow2', '-F', 'qcow2', '-b', source_image, target_image)
+        else:
+            extra_args = settings.extra_args('cp')
+            cp(*extra_args, source_image, target_image)
 
         # Setup virt-sysprep args.
         if settings.dns_domain:
@@ -702,7 +803,7 @@ class Instance:
         """
         self.start()
         self.meta.pop('address', None) # Flush our cached address.
-        address = self.address() # Wait until booted.
+        address = self.address()
         user = self.meta['user']['username']
         ssh_identity = self.meta['user']['ssh_identity']
         args = [
@@ -714,6 +815,8 @@ class Instance:
             '-o', 'LogLevel=ERROR',
             f'{user}@{address}',
         ]
+
+        self.wait_for_port(22)  # Wait until ssh port is ready.
         if not command:
             args.insert(0, ssh.__name__) # Required for execv.
             os.execv(ssh.__name__, args) # Drop into interactive shell, never to return.
